/*
 This file is part of the digitalSTROM ds485-client library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-client library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM ds485-client library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM ds485-client library. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdint.h>
#include <signal.h>
#include <setjmp.h>
#include <syslog.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <string.h>
#include <sys/time.h>

#include <digitalSTROM/ds485.h>
#include <digitalSTROM/ds485-socket-interface.h>
#include "ds485-client-interface.h"
#include "ds485-client.h"

#ifndef SOCK_CLOEXEC
#define SOCK_CLOEXEC 0
#endif

/******************************************************************************
 * Local Data
 */
static ds485_client_interface_t connInterfaceSocket =
{
        .conntype = CLIENT_SOCKET,
        .open = ds485_socket_client_open,
        .close = ds485_socket_client_close,
        .send = ds485_socket_client_send_command,
        .get_bus_devices = ds485_socket_client_get_bus_devices,
        .get_bus_state = ds485_socket_client_get_bus_state,
        .get_dsid = ds485_socket_client_get_dsid,
        .restart_bus = ds485_socket_client_restart_bus,
};

static int connectsock(const char *host, const char *service, const char *protocol, uint8_t *active);
static int sockwrite(int *fd, uint8_t* data, size_t dlen);
static int socket_service_dispatch(ds485_client_t *cl, uint8_t* data, int dlen);
static int socket_connect(ds485_client_t *cl);
static void* socket_thread(void* arg);
static void socket_cleanup(void *arg);


#ifdef DEBUG
#define debugmessage printf
#else
#define debugmessage(...)
#endif

/**
 * Initialize socket client data
 */
int ds485_socket_client_init(struct _ds485_client_s *cl)
{
    const char *protocol, *name, *port;
    char *connection = strdup(cl->connspec);

    protocol = strtok(connection, ":");
    if (protocol == NULL) {
        return ERROR_INVALID_CONNSPEC;
    }
    if (!strcmp(protocol, "tcp")
#if 0
            || !strcmp(protocol, "udp")
#endif
            ) {
        char *name_port = strdup(cl->connspec + 6);
        name = strtok(name_port, ":");
        if (name == NULL) {
            return ERROR_INVALID_CONNSPEC;
        }
        port = strtok(NULL, ":");
        if (port == NULL) {
            return ERROR_INVALID_CONNSPEC;
        }
        cl->socketClient.dProtocol = strdup(protocol);
        cl->socketClient.dName = strdup(name);
        cl->socketClient.dPort = strdup(port);
        cl->connInterface = connInterfaceSocket;
        free(name_port);
    }
#if 0
    else if(!strcmp(protocol, "unix")) {
        char *name_port = strdup(cl->connspec + 7);
        name = strtok(name_port, ":");
        if (name == NULL) {
            return ERROR_INVALID_CONNSPEC;
        }
        cl->socketClient.dProtocol = strdup(protocol);
        cl->socketClient.dName = strdup(name);
        cl->connInterface = connInterfaceSocket;
        cl->connInterface.conntype = CLIENT_UNIX;
        free(name_port);
    }
#endif
    else {
        return ERROR_INVALID_CONNSPEC;
    }
    cl->socketClient.sockfd = -1;
    free(connection);

    return 0;
}

int ds485_socket_client_open(struct _ds485_client_s *cl, const dsid_t *dsid)
{
    /* Start communication thread */
    if (pthread_create(&cl->connThread, NULL, socket_thread, cl)) {
        return ERROR_THREAD_START_FAILED;
    }
    pthread_detach(cl->connThread);

    return 0;
}

/**
 * Closes the session, frees all connection related ressources
 */
int ds485_socket_client_close(struct _ds485_client_s *cl)
{
    cl->active = 0;
    pthread_join(cl->connThread, NULL);
    return 0;
}

int ds485_socket_client_get_bus_state(ds485_client_t *cl, bus_state_t *state)
{
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_QUERY_BUSSTATE;
    cmdCall.length = 3;

    return sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmdCall, cmdCall.length);
}

int ds485_socket_client_get_dsid(ds485_client_t *cl)
{
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_QUERY_DSID;
    cmdCall.length = 3;

    return sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmdCall, cmdCall.length);
}

int ds485_socket_client_restart_bus(ds485_client_t *cl)
{
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_RESTART_BUS;
    cmdCall.length = 3;

    return sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmdCall, cmdCall.length);
}

int ds485_socket_client_get_bus_devices(ds485_client_t *cl, dsid_t *device_list, int device_list_size)
{
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_QUERY_DEVICES;
    cmdCall.length = 4;
    cmdCall.data[0] = device_list_size;

    return sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmdCall, cmdCall.length);
}

/**
 * Send dM-API command
 */
int ds485_socket_client_send_command(struct _ds485_client_s *cl, ds485_container_t *request)
{
    dS485CmdApidata cmdCall;

    memset(&cmdCall, 0, sizeof(dS485CmdApidata));
    cmdCall.apiData.destinationId = request->destinationId;
    cmdCall.apiData.sourceId = request->sourceId;
    cmdCall.apiData.containerType = request->containerType;
    cmdCall.apiData.length = request->length;
    memcpy(&cmdCall.apiData.data, request->data, cmdCall.apiData.length);

    cmdCall.length = DS485_CONTAINER_SIZE + cmdCall.apiData.length + 3;
    cmdCall.command = DS485D_CMD_TX_PACKET;

    return sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmdCall, cmdCall.length);
}

static int ds485_socket_client_ping(ds485_client_t *cl)
{
    dS485dCommand cmdCall;

    cmdCall.command = DS485D_CMD_PING;
    cmdCall.length = 0;
    cmdCall.data[0] = 0;

    return sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmdCall, cmdCall.length);
}

/************************ SOCKET CONNECTION *******************************/

static int connectsock(const char *host, const char *service, const char *protocol, uint8_t *active)
{
    struct hostent *phe;
    struct servent *pse;
    struct protoent *ppe;
    struct sockaddr_in sin;
    int s, type, proto, opts, one = 1;
    struct timeval tv;
    fd_set wdfd;

    memset((char *) &sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;

    if ((pse = getservbyname(service, protocol)))
        sin.sin_port = htons(ntohs((unsigned short) pse->s_port));
    else if ((sin.sin_port = htons((unsigned short) atoi(service))) == 0)
        return ERROR_NO_PORT_FOUND;

    if ((phe = gethostbyname(host)))
        memcpy((char *) &sin.sin_addr, phe->h_addr, phe->h_length);
    else if ((sin.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE)
        return ERROR_INVALID_ADDRESS;

    ppe = getprotobyname(protocol);
    if (strcmp(protocol, "udp") == 0) {
        type = SOCK_DGRAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_UDP;
    }
    else {
        type = SOCK_STREAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_TCP;
    }

    if ((s = socket(PF_INET, SOCK_CLOEXEC | type, proto)) < 0)
        return ERROR_CREATE_SOCKET;

#ifdef __APPLE__
    fcntl(s, F_SETFD, FD_CLOEXEC);
#endif

    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof(one)) == -1) {
        close(s);
        return ERROR_SOCKET_OPTIONS;
    }

    opts = fcntl(s, F_GETFL);
    if (opts > 0) {
        fcntl(s, F_SETFL, opts | O_NONBLOCK);
    }

    FD_ZERO(&wdfd);
    FD_SET(s, &wdfd);

    int ret = connect(s, (struct sockaddr *) &sin, sizeof(sin));

    if (ret != 0) {
        if (errno != EINPROGRESS && errno != EALREADY) {
            perror("connectsock connect:");
            close(s);
            return ERROR_SOCKET_CONNECT;
        }

        /* socket is non-blocking and we have to poll for connect completion.
         *
         * Non-blocking mode is used to work around very long (>180s) timeouts
         * that result in the process not being interruptible if there is no
         * server available. */
        while (*active) {
            int fds;
            tv.tv_sec = 1;
            tv.tv_usec = 0;

            fds = select(s + 1, NULL, &wdfd, NULL, &tv);
            if (fds == 0) {
                /* timeout */
                continue;
            } else if (fds < 0) {
                perror("connectsock select:");
                close(s);
                return ERROR_SOCKET_CONNECT;
            }

            /* see "man connect" EINPROGRESS */
            int err = -1;
            int l_sock_optval_len = sizeof(err);
            if (getsockopt(s, SOL_SOCKET, SO_ERROR, (int*)&err,
                           (socklen_t*)&l_sock_optval_len) != 0) {
                perror("connectsock getsockopt:");
                close(s);
                return ERROR_SOCKET_CONNECT;
            }

            if (err != 0) {
                fprintf(stderr, "connectsock einprogress: %s\n", strerror(err));
                close(s);
                return ERROR_SOCKET_CONNECT;
            }

            break;
        }
    }

    /* return to blocking mode */
    opts = fcntl(s, F_GETFL);
    if (opts > 0) {
        fcntl(s, F_SETFL, opts & ~O_NONBLOCK);
    }

#ifdef IPTOS_LOWDELAY
    opts = IPTOS_LOWDELAY;
    setsockopt(s, IPPROTO_IP, IP_TOS, &opt, sizeof opt);
#endif

#ifdef TCP_NODELAY
    if (type == SOCK_STREAM)
        setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &one, sizeof one);
#endif
    return s;
}

static int sockwrite(int *fd, uint8_t* data, size_t dlen)
{
    int status = write(*fd, data, dlen);

    if (status == dlen) {
        return 0;
    }
    else if (status > -1) {
        debugmessage("short write disconnecting socket[%d]\n", *fd);
        close(*fd);
        *fd = -1;
        return ERROR_SOCKET_FULL;
    }
    else if (errno == EAGAIN || errno == EINTR) {
        debugmessage("socket[%d] write failed, no data sent.\n", *fd);
        return ERROR_SOCKET_FAILED;
    }
    else if (errno == EBADF) {
        status = ERROR_SOCKET_VANISHED;
        debugmessage("socket[%d] has vanished.\n", *fd);
    }
    else if (errno == EWOULDBLOCK /*&& (timestamp() - cl->active) > NOREAD_TIMEOUT*/) {
        status = ERROR_SOCKET_WOULD_BLOCK;
        debugmessage("socket[%d] timed out.\n", *fd);
    }
    else {
        status = ERROR_SOCKET_UNKNOWN;
        debugmessage("socket[%d] write: %s\n", *fd, strerror(errno));
    }
    close(*fd);
    *fd = -1;
    return status;
}

/******************************* RECEIVE THREAD *******************************/

static int socket_service_dispatch(ds485_client_t *cl, uint8_t* data, int dlen)
{
    ds485_container_t container;
    dS485dCommand* cmd;
    uint16_t len;

    cmd = (dS485dCommand *) data;
    len = cmd->length;

    switch (cmd->command)
    {

    case DS485D_CMD_API_REVISION:
        debugmessage("Server API Revision: %d\n", cmd->data[0]);
        break;

    case DS485D_CMD_PING:
        break;

    case DS485D_CMD_TX_PACKET:
        break;

    case DS485D_CMD_RX_PACKET:
    {
        ds485_container_t* pData = (ds485_container_t *) cmd->data;

        /* copy data */
        container = *pData;

        /* waiting synchronous commands */
        pthread_mutex_lock(&cl->syncSend.mutex);
        if (cl->syncSend.state == STATE_WAITING) {
            if (IsEqualId(cl->syncSend.response.sourceId, container.sourceId) &&
                (cl->syncSend.response.containerType == container.containerType) &&
                (cl->syncSend.response.data[0] == container.data[0]) &&
                (cl->syncSend.response.data[1] == container.data[1]))
            {
                cl->syncSend.response = container;
                cl->syncSend.state = STATE_RECEIVED;
                pthread_cond_signal(&cl->syncSend.condition);
                pthread_mutex_unlock(&cl->syncSend.mutex);
                break;
            }
        }
        pthread_mutex_unlock(&cl->syncSend.mutex);

        /* Receive Callback */
        if (cl->connInterface.responseCallback) {
            cl->connInterface.responseCallback(cl->connInterface.responseCallbackData, &container);
        }

        break;
    }

    case DS485D_CMD_DEVICE_PACKET:
        break;

    case DS485D_CMD_QUERY_BUSSTATE:
    {
        pthread_mutex_lock(&cl->busQuery.mutex);
        cl->busQuery.length = cmd->data[0];
        cl->busQuery.state = STATE_RECEIVED;
        pthread_cond_signal(&cl->busQuery.condition);
        pthread_mutex_unlock(&cl->busQuery.mutex);
        break;
    }

    case DS485D_CMD_QUERY_DEVICES:
    {
        pthread_mutex_lock(&cl->busQuery.mutex);
        if (cl->busQuery.state == STATE_WAITING)
        {
            cl->busQuery.length = (len - 2) / sizeof(dsid_t);
            memcpy(cl->busQuery.devices, cmd->data, cl->busQuery.length * sizeof(dsid_t));
            cl->busQuery.state = STATE_RECEIVED;
        }
        pthread_cond_signal(&cl->busQuery.condition);
        pthread_mutex_unlock(&cl->busQuery.mutex);
        break;
    }

    case DS485D_CMD_BUS_CHANGE:
    {
        ds485_container_t* pData = (ds485_container_t *) cmd->data;
        if (cl->connInterface.busChangeCallback != NULL)
        {
            cl->connInterface.busChangeCallback(cl->connInterface.busChangeCallbackData, &pData->sourceId, pData->data[0]);
        }
        break;
    }

    case DS485D_CMD_BUS_STATE:
    {
        ds485_container_t* pData = (ds485_container_t *) cmd->data;
        cl->busState = (bus_state_t) pData->data[0];
        if (cl->connInterface.stateCallback != NULL)
        {
            cl->connInterface.stateCallback(cl->connInterface.stateCallbackData, cl->busState);
        }
        break;
    }

    case DS485D_CMD_QUERY_DSID:
    {
        pthread_mutex_lock(&cl->busQuery.mutex);
        if (cl->busQuery.state == STATE_WAITING)
        {
            memcpy(cl->busQuery.devices, cmd->data, sizeof(dsid_t));
            cl->busQuery.state = STATE_RECEIVED;
        }
        pthread_cond_signal(&cl->busQuery.condition);
        pthread_mutex_unlock(&cl->busQuery.mutex);
        break;
    }

    case DS485D_CMD_RESTART_BUS:
    {
        pthread_mutex_lock(&cl->busQuery.mutex);
        if (cl->busQuery.state == STATE_WAITING)
        {
            cl->busQuery.state = STATE_RECEIVED;
        }
        pthread_cond_signal(&cl->busQuery.condition);
        pthread_mutex_unlock(&cl->busQuery.mutex);
        break;
    }

    case DS485D_CMD_REGISTER:
    case DS485D_CMD_UNREGISTER:
    case DS485D_CMD_PACKETFILTER:
    case DS485D_CMD_RAWMODE:
    case DS485D_CMD_QUERY_ADDRESS:
        debugmessage("invalid server command\n");
        break;

    default:
        debugmessage("unknown server command\n");
        break;
    }

    return len;
}

static int socket_connect(ds485_client_t *cl)
{
    int rval;

    cl->socketClient.sockfd = connectsock(cl->socketClient.dName, cl->socketClient.dPort,
            cl->socketClient.dProtocol, &cl->active);

    if (cl->socketClient.sockfd < 0) {
        cl->busState = DS485_STATE_ISOLATED;
        return ERROR_INVALID_FD;
    }
    //debugmessage("socket_connect: connected to ...);

    /* Set connection options */
    dS485dCommand cmd;
    cmd.command = DS485D_CMD_PACKETFILTER;
    cmd.length = 4;
    cmd.data[0] = (cl->flags & PROMISCUOUS_MODE);
    rval = sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmd, cmd.length);

    /* query bus state */
    cmd.command = DS485D_CMD_BUS_STATE;
    cmd.length = 3;
    rval = sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmd, cmd.length);

    return rval;
}

static void socket_cleanup(void *arg)
{
    ds485_client_t *cl = (ds485_client_t *) arg;

    free(cl->socketClient.dName);
    free(cl->socketClient.dPort);
    free(cl->socketClient.dProtocol);

    ds485_cleanup(cl);
}

static void* socket_thread(void* arg)
{
    ds485_client_t *cl = (ds485_client_t *) arg;
    struct timeval tv;
    fd_set rfds;
    int rval;
    int rxLen = 0;
    int idleTimer = 0;

#define RXBUFLEN 1024
    uint8_t* rxBuffer = (uint8_t *) malloc(RXBUFLEN);
    uint8_t* p;

    pthread_cleanup_push(&socket_cleanup, (void*)cl);

    while (cl->active) {

        /* connect */
        if (cl->socketClient.sockfd < 0) {

            /* first time running through reconnect, send isolated state to client */
            if (cl->busState == DS485_STATE_DISCONNECTED) {
                if (cl->connInterface.stateCallback != NULL) {
                    cl->connInterface.stateCallback(cl->connInterface.stateCallbackData, DS485_STATE_DISCONNECTED);
                }
                cl->busState = DS485_STATE_ISOLATED;
            }

            rval = socket_connect(cl);
            if (rval < 0) {
                sleep(1);
                continue;
            }

            /* force a callback to be generated with the current bus state */
            {
                dS485dCommand cmdCall;
                cmdCall.command = DS485D_CMD_BUS_STATE;
                cmdCall.length = 3;
                sockwrite(&cl->socketClient.sockfd, (uint8_t *) &cmdCall, cmdCall.length);
            }
        }

        pthread_mutex_lock(&cl->openQuery.mutex);
        if (cl->openQuery.state == STATE_IDLE || cl->openQuery.state == STATE_RECEIVED)
        {
            /* open took too long */
            shutdown(cl->socketClient.sockfd, 0);
            close(cl->socketClient.sockfd);
            cl->socketClient.sockfd = -1;
            cl->active = 0;
        }
        else if (cl->openQuery.state == STATE_WAITING)
        {
            cl->openQuery.state = STATE_RECEIVED;
            pthread_cond_signal(&cl->openQuery.condition);
        }
        else /* cl->openQuery.state == STATE_CONNECTED */
        {
            /* automatic reconnect */
        }
        pthread_mutex_unlock(&cl->openQuery.mutex);

        /* as long as the socket is connected */
        while (cl->socketClient.sockfd > 0 && cl->active) {
            FD_ZERO(&rfds);
            FD_SET(cl->socketClient.sockfd, &rfds);

            tv.tv_sec = 1;
            tv.tv_usec = 0;
            rval = select(cl->socketClient.sockfd + 1, &rfds, NULL, NULL, &tv);
            if ((rval < 0) || (cl->active == 0)) {
                if (errno == EINTR || errno == EINPROGRESS)
                    continue;
                debugmessage("client socket select error %d/%s [fd%d]\n", errno,
                             strerror(errno), cl->socketClient.sockfd);
                continue;
            } else if (rval == 0) {
                if (rxLen != 0) {
                    debugmessage("client socket timeout: reset rxbuffer (prev: %d) [%d]\n", rxLen, cl->socketClient.sockfd);
                }
                rxLen = 0;
                if (idleTimer ++ > 60) {
                    if ((rval = ds485_socket_client_ping(cl)) < 0) {
                        debugmessage("client socket ping error %d [%d]\n",
                                     rval, cl->socketClient.sockfd);
                        goto disconnect;
                    }
                    idleTimer = 0;
                }
            } else {
                if (FD_ISSET(cl->socketClient.sockfd, &rfds)) {
                    rval = read(cl->socketClient.sockfd, rxBuffer + rxLen, RXBUFLEN - rxLen);
                    if (rval < 0) {
                        debugmessage("client socket read error %d [%d]\n", rval, cl->socketClient.sockfd);
                        continue;
                    }
                    else if (rval == 0) {
                        debugmessage("client socket closed by server [%d]\n",
                                     cl->socketClient.sockfd);
                        goto disconnect;
                    }
                    idleTimer = 0;

                    rxLen += rval;
                    //if (verbose) printf("sockfd %d, rx %d bytes, total %d\n", cl->socketClient.sockfd, rval, rxLen);

                    p = rxBuffer;
                    while (rxLen >= 3) {
                        /* sanity checks */
                        if (p[0] == 0 && p[1] == 0 && p[2] == 0) {
                            rxLen = 0;
                            debugmessage("buffer resync [%d]\n", cl->socketClient.sockfd);
                            break;
                        }

                        /* not enough data */
                        if ((p[2] << 8 | p[1]) > rxLen)
                            break;

                        /* Process packet data ... */
                        rval = socket_service_dispatch(cl, p, rxLen);

                        p += rval;
                        rxLen -= rval;
                    }
                    if (p != rxBuffer && rxLen > 0) {
                        memcpy(rxBuffer, p, rxLen);
                    }

                }
            }
        }
disconnect:
        shutdown(cl->socketClient.sockfd, SHUT_RDWR);
        close(cl->socketClient.sockfd);
        cl->socketClient.sockfd = -1;
        cl->busState = DS485_STATE_DISCONNECTED;
    }

    free(rxBuffer);
    pthread_cleanup_pop(1);
    return NULL;
}
